# Copyright (c) 2001-2005, RedIRIS. All Rights Reserved.
#
# You may distribute under the terms of the GNU General Public License,
# as specified in the LICENSE file that was shipped with this distribution.

package PAPI::PoA;

use MIME::Base64 qw/decode_base64 encode_base64/;
use URI::Escape;
use Crypt::OpenSSL::RSA;
use Crypt::CBC;
use Crypt::Rijndael;
use constant PADDINGSIZE => 11;
use Fcntl;
use DB_File;
use DB_File::Lock;
use Data::Dumper;
use Digest::MD5 qw(md5_base64);
use PAPI::SPOCP;
use PAPI::Config;
use strict;

sub new
{
	my $object={@_};
	$object->{'Id'}=$$ . "_" . time;
	return bless $object;
}

sub set_parameters
{
	my $self=shift;
	my $servername=shift;
	my $serv_id=shift;
	
	my $serverAux = $PAPI::Config::Global->{$servername};
	my $lconf;
	my $gconf;
    
 	$self->accesslog("PoA::set_parameters# Servername = <$servername>\n");
 	$self->accesslog("PoA::set_parameters# Server_ID = <$serv_id>\n");       
      
	if ( (ref($serverAux->{$serv_id}) ne "HASH") || ($serverAux->{$serv_id}{Serv} ne $serv_id) ){
    	$self->error("Apache Server_ID = <".$serv_id.">");
	    $self->error("PAPI XML Service_ID = <".$serverAux->{$serv_id}{Serv}.">");
    	$self->error("Can't instance a new PoA, the Service_ID from Apache conf. is not equal to the id for the Location from PAPI XML configuration file.");
	    return 0;
	}
  
  	$gconf = $PAPI::Config::Global->{$servername}{$serv_id};
	
	%{ $self} = (%{ $self}, %{ $gconf});
	
	$self->{'registerKey'}='';
	$self->{'registerVal'}='';
	$self->{'rawHcook'}='';
	$self->{'attrList'}= [];

	########### Checking parameters ########################
	if ($self->{'Hkey'} eq ""){
		$self->error("Hkey parameter is empty");
		return 0;
	}
	if ($self->{'Lkey'} eq ""){
		$self->error("Lkey parameter is empty");
		return 0;
	}
	if ($self->{'Pubkeys_Path'} eq ""){
		$self->error("Pubkeys_Path parameter is empty");
		return 0;
	}
	if ($self->{'Serv'} eq ""){
		$self->error("Service_ID parameter is empty");
		return 0;
	}
	if ($self->{'Hcook_DB'} eq ""){
		$self->error("Hcook_DB parameter is empty");
		return 0;
	}
	if ($self->{'Accept_File'} eq ""){
		$self->error("Accept_File parameter is empty");
		return 0;
	}
	if ($self->{'Reject_File'} eq ""){
		$self->error("Reject_File parameter is empty");
		return 0;
	}
	if (($self->{'Req_DB'} eq "") && ($self->{'GPoA_URL'} ne "")){
                $self->error("Req_DB parameter is empty");
                return 0;
        }

	$self->{'MxNonceErr'}=3 unless ($self->{'MxNonceErr'} ne "");
	$self->{'Auth_Location'}="/PAPI/cookie_handler.cgi" unless ($self->{'Auth_Location'} ne  "");
	$self->{'Hcook_Handler'}=undef unless ($self->{'Hcook_Handler'} ne  "");
	$self->{'Lcook_Timeout'}=60 unless ($self->{'Lcook_Timeout'} ne  "");
	$self->{'CRC_Timeout'}=30 unless ($self->{'CRC_Timeout'} ne  "");
	$self->{'URL_Timeout'}=90 unless ($self->{'URL_Timeout'} ne  "");
	$self->{'Debug'}=0 unless ($self->{'Debug'} ne "");
	$self->{'Id'}.="_".$self->{'Serv'};
	$self->{'SPOCPAcc'}=0 unless ($self->{'SPOCPAcc'} ne "");
	$self->{'SPOCPAct'}="accept" unless ($self->{'SPOCPAct'} ne "");
	$self->{'PxCkSize'}=320768 unless ($self->{'PxCkSize'} ne "");

	return 1;
}

sub test_filters
{
	my $self=shift;
	my $str=shift;

# Build the attribute list
	$self->build_attrList ($str);
# PAPI_Filter directives are evaluated first
	for my $filter (@{$self->{filters}}) {
		## Each filter is a hash with the keys "reg" for the regex and "act" for the action
		my $s = $filter->{'reg'};
		my $v = $filter->{'act'};
		if($str =~ /$s/){
			if (defined $v && $v eq "accept") {
				$self->accesslog("Filter /$s/ matches for $str: accept");
				return 1;
			}
			else {
				$self->accesslog("Filter /$s/ matches for $str: reject");
				return 0;
			}
		}
	}
# If SPOCP use is defined, ask the server
	if (defined $self->{'SPOCPSrv'}) {
		my $spocp=new PAPI::SPOCP($self->{'SPOCPSrv'});
		if (!defined $spocp) {
			$self->error("Unable to connect to SPOCP server");
			return $self->{'SPOCPAcc'};
		}
		my $rt = $spocp->query($self->{'Serv'},$self->{'SPOCPAct'},$self->{'attrList'});
		$self->accesslog("Querying SPOCP server for \"$str\". Result: $rt");
		return $rt;
	}
	return 1;
}

sub build_attrList
{
	my $self=shift;
	my $str=shift;

	my $are='\s*,\s*';
	if (defined $self->{'AttSep'} &&  $self->{'AttSep'} ne "") {
		$are='\s*'.$self->{'AttSep'}.'\s*';
	}
	my $ave='\s*=\s*';
	if (defined $self->{'ValSep'} &&  $self->{'ValSep'} ne "") {
		$ave='\s*'.$self->{'ValSep'}.'\s*';
	}
	my @acomp=split (/@/,$str);
	my $asid=pop(@acomp);
	my $spsubj=join('@',@acomp);
	for my $atc (split (/$are/,$spsubj)) {
		push (@{$self->{'attrList'}}, split (/$ave/,$atc,2));
	}
	push (@{$self->{'attrList'}},("PAPIAuthServer",$asid));
}

sub test_AS
{
	my $self=shift;
	my $str=shift;

	$self->debug("PoA# test_AS: Testing str AS:<".$self->{ases}{$str}.">") if ($self->{'Debug'});
	return (defined($self->{ases}{$str}));
}
		
sub error
{
	my $self=shift;
	
	my $str=shift;
	
	print STDERR "PAPI:*ERROR*:#$self->{'Id'}: $str\n";
}

sub warning
{
	my $self=shift;
	my $str=shift;

	print STDERR "PAPI:*WARN*:#$self->{'Id'}: $str\n";
}

sub accesslog
{
	my $self=shift;
	my $str=shift;

	print STDERR "PAPI:#$self->{'Id'}:$str\n";
}

sub debug
{
	my $self=shift;
	my $str=shift;
 
	print STDERR "PAPI:-->Debug:#$self->{'Id'}:$str\n";
} 

######## Verifying signed URLs ####################
#
#	Function: 	test_URL
#	Parameters:	str,file_name,type_key
#		
#	    str:	DATA field of the URL
#	    file_name:	file of the public key
#	    type_key:	format of the public key file
#		FORMAT_UNDEF    0
#		FORMAT_ASN1     1
#		FORMAT_TEXT     2
#		FORMAT_PEM      3
#		FORMAT_NETSCAPE 4
#		FORMAT_PKCS12   5
#		FORMAT_SMIME    6
#
####################################################

sub test_URL
{
	my $self=shift;

	my $as=uri_unescape($_[0]);
	my $data=uri_unescape($_[1]);

	if(!$self->test_AS($as)){
		$self->warning("Authentication server $as not defined for this PoA");
		return (0,0);
	}
	my ($key_file,$key_type)=$self->pub_keyfile($as);
	$key_file=uri_unescape($key_file);
	my $linea = rsaDecrypt($data, $key_file);
	
	if($linea eq "ERRORFILE"){
		$self->error("Error reading public key file:$key_file of type:$key_type");
		return (0,0);
	}
	if($linea eq "ERRORCRYPT"){
		$self->warning("Error in verifying signed URL");
		return (0,0);
	}
	my @elem= ("",0,0,"","");
	my @splin=split(/:/,$linea);
    $elem[4]=pop @splin; $elem[3]=pop @splin;
    $elem[2]=pop @splin; $elem[1]=pop @splin;
    $elem[0]=join(':',@splin);
        
	my $now=time;
	if ($now > $elem[2] + $self->{'URL_Timeout'}){
		$self->warning("URL timeout expired: $now > $elem[2] + $self->{'URL_Timeout'}");
		return (0,0);
	}
	if($now > $elem[1]){
		$self->warning("Validation date field in URL is wrong\n");
		return (0,0);
	}
	if ($elem[4] ne $self->{Serv}) {
		$self->warning("Invalid service ID $elem[4] in URL");
		return (0,0);
	}
	if (!$self->test_filters("$elem[0]\@$as")) { return (0,0); }
	my $rt="\@$as";
	if (defined $self->{RwUser}) {
		my $nua=$elem[0];
		my $evs="\$nua =~ s($self->{RwUser}->{regex})($self->{RwUser}->{value})gsx";
		eval $evs;
		$self->accesslog("User assertion $elem[0] transformed into $nua");
		$elem[0]=$nua;
	}
	if (defined $self->{HashUser} && $self->{HashUser} eq "1") {
		my $nua=md5_base64($elem[0]);
		$self->accesslog("User assertion $elem[0] transformed into $nua");
		$elem[0]=$nua;
	}
	if (defined $self->{HcookGen}) {
		no strict;
		my @pkgc=split (/::/,$self->{HcookGen});
		pop(@pkgc);
		my $pkgn=join('/',@pkgc);
		require "$pkgn.pm";
		$dref=$self->{HcookGen};
		my $nua=&$dref($elem[0],$as);
		$self->accesslog("User assertion $nua generated from $elem[0] by ".$self->{HcookGen});
		$self->{'attrList'} = [];
		$elem[0]=$nua;
	}
	$rt=$elem[0].$rt;
	if (defined $self->{MaxTtl}) {
		my $maxt=$elem[2]+$self->{MaxTtl};
		if ($elem[1] > $maxt) {
			$self->warning("Requested TTL of $elem[1] reduced to $maxt ($elem[2] + $self->{MaxTtl})");
			$elem[1]=$maxt;
		}
	}
	return ($rt,$elem[1]);
}

sub test_GPoA_URL
{
	my $self=shift;
	my $datarg=shift;
	my $gid=shift;
	if (!defined $gid || $gid eq "") { 
		$gid="_GPoA"; 
	}
	if (!defined $datarg || $datarg eq ""){
		$self->error("Signed GPoA/AS URL without DATA field");
		return (0,0,0);
	}
	my ($key_file,$key_type)=$self->pub_keyfile($gid);
	$key_file=uri_unescape($key_file);
	my $data=uri_unescape($datarg);
	my $linea = rsaDecrypt($data, $key_file);

	if($linea eq "ERRORFILE"){
		$self->error("Error reading public key file:$key_file of type:$key_type");
		return (0,0,0);
	}
	if($linea eq "ERRORCRYPT"){
		$self->warning("Error in verifying signed URL from GPoA/AS");
		return (0,0,0);
	}
	$self->debug("Decrypted GPoA_URL: $linea") if ($self->{'Debug'});
	my @elem= ("",0,0,"");
	my @splin=split(/:/,$linea);
    $elem[3]=pop @splin; $elem[2]=pop @splin; $elem[1]=pop @splin;
    $elem[0]=join(':',@splin);

	my $now=time;
	if($elem[3] eq ""){
		$self->warning("Signed GPoA/AS URL without Id_Req");
		return (0,0,0);
	}
	if ($elem[0] eq "ERROR"){
		my $lv = $self->{'GPoA_URL'};
		$self->warning("Authentication ERROR received from GPoA $lv");
		return ($elem[3],0,0);
	}
	if ($now > $elem[2] + $self->{'URL_Timeout'}){
		my $lv = $elem[2] + $self->{'URL_Timeout'};
		$self->warning("GPoA/AS URL timeout expired: $now > $lv");
		return ($elem[3],0,0);
	}
	if($now > $elem[1]){
		$self->warning("Validation date field in GPoA/AS URL is wrong");
		return ($elem[3],0,0);
	}
	if (!$self->test_filters("$elem[0]")) { 
		return ($elem[3],0,0); 
	}
	my @ccomp=split (/@/,$elem[0]);
	my $as=pop(@ccomp);
	my $rt='@'.$as;
	$elem[0]=join('@',@ccomp);
	if (defined $self->{RwUser}) {
		my $nua=$elem[0];
		my $evs="\$nua =~ s($self->{RwUser}->{regex})($self->{RwUser}->{value})gsx";
		eval $evs;
		$self->accesslog("User assertion $elem[0] transformed into $nua");
		$elem[0]=$nua;
	}
	if (defined $self->{HashUser} && $self->{HashUser} eq "1") {
		my $nua=md5_base64($elem[0]);
		$self->accesslog("User assertion $elem[0] transformed into $nua");
		$elem[0]=$nua;
	}
	if (defined $self->{HcookGen}) {
		no strict;
		my @pkgc=split (/::/,$self->{HcookGen});
		pop(@pkgc);
		my $pkgn=join('/',@pkgc);
		require "$pkgn.pm";
		$dref=$self->{HcookGen};
		my $nua=&$dref($elem[0],$as);
		$self->accesslog("User assertion $nua generated from $elem[0] by ".$self->{HcookGen});
		$self->{'attrList'} = [];
		$elem[0]=$nua;
	}
	$rt=$elem[0].$rt;
	if (defined $self->{MaxTtl}) {
		my $maxt=$elem[2]+$self->{MaxTtl};
		if ($elem[1] > $maxt) {
			$self->warning("Requested TTL of $elem[1] reduced to $maxt ($elem[2] + $self->{MaxTtl})");
			$elem[1]=$maxt;
		}
	}
	return ($elem[3],$rt,$elem[1]);
}


###### Read configuration parameters for a PoA ############
#
#	Function: 	read_key
#	Parameters:	file_name
#		
#	    str:

#sub read_key
#{
#	my $self=shift;
# 
#        open(KEY,"$_[0]");
#        my $res=<KEY>;
#        chop($res);
#        close(KEY);
#        return $res;
#} 

sub pub_keyfile
{
	my $self=shift;
	my $name=shift;
	
	return ($self->{'Pubkeys_Path'} . "/" . $name . "_pubkey.pem",3);
}



###### Management of cookies ######################

sub get_cookies
{
	my $self=shift;
	my $cookies=shift;

	my ($Lcook) = $cookies =~ /$self->{Lcook_Name}=([^;]+)/;
	$Lcook=uri_unescape($Lcook);
	my ($Hcook) = $cookies =~ /$self->{Hcook_Name}=([^;]+)/;
	$Hcook=uri_unescape($Hcook);
	return ($Hcook,$Lcook);
}

sub gen_Hcook
{
	my $self=shift;
	my $code=shift;
	my $valid_date=shift;
	my $temporal=shift;
	my $key=shift;
	my $crc=shift;
	
	if(!defined($crc)){
# $crc is not a cryptographically secure random number, but its randomness
# fits our purposes
		srand;
		my $now = time;
		$crc=int (10000 * rand ($now));
		$self->{'registerKey'} = "$key:$self->{'Loc'}:$self->{'Serv'}";
		$self->{'registerVal'} = "0:$crc:$valid_date:$now:0";
		$self->reg_Hcook();
	}
	$self->debug("Encrypting: #$temporal:$code:$valid_date:$crc:$self->{'Loc'}:$self->{'Serv'}#") if ($self->{'Debug'});

	my $iv = pack("LLLL",rand(0xffffffff),rand(0xffffffff),rand(0xffffffff),rand(0xffffffff));
	$self->debug("Hkey: ".$self->{'Hkey'}) if ($self->{'Debug'});

	my $key_aes=$self->{'Hkey'};
	my $dec_obj = new Crypt::CBC({key=>$key_aes,cipher=>'Rijndael',iv=>$iv,regenerate_key=>0,padding=>'null',prepend_iv=>0});
	my $Hcook = encode_base64($iv.$dec_obj->encrypt("$temporal:$code:$valid_date:$crc:$self->{'Loc'}:$self->{'Serv'}"));
	$self->debug("Encrypted: #$Hcook#") if ($self->{'Debug'});
	$self->{'rawHcook'}=$Hcook;
	$Hcook=uri_escape($Hcook);

	if ($temporal){
		return "$self->{Hcook_Name}=$Hcook; path=$self->{'Loc'}; domain=$self->{'Domain'}";
	}
	
	my $gmnow=gmtime($valid_date);
	my @com=split(/\s+/,$gmnow); 
	my $max_age="$com[0], $com[2]-$com[1]-$com[4] $com[3] GMT";

	return "$self->{Hcook_Name}=$Hcook; expires=$max_age; path=$self->{'Loc'}; domain=$self->{'Domain'}";
}

sub test_Hcook
{
	my $self=shift;
	my $Hcook=shift;
	my $gen_crc=shift;

	my $now=time;
	
	$Hcook=uri_unescape($Hcook);
 
	if ($Hcook eq ""){
		$self->accesslog("Hcook is empty");
		return (0,0,1,0,0);
	}
	$self->debug("Decrypting: #$Hcook#") if ($self->{'Debug'});
 
	my $enc_packet=decode_base64($Hcook);
    my $iv = substr($enc_packet,0,16);
    $enc_packet= substr($enc_packet,16);
    my $key_aes=$self->{'Hkey'};
    my $dec_obj = new Crypt::CBC({ key=>$key_aes,
				   cipher=>'Rijndael',
				   iv=>$iv,
                    regenerate_key=>0,
				   padding=>'null',
				   prepend_iv=>0
				   });
	my $Hcook_des=$dec_obj->decrypt($enc_packet);

	$self->debug("Hcook_des=$Hcook_des") if ($self->{'Debug'}); 
	if ($Hcook_des eq "ERRORCRYPT"){
		$self->warning("Error decrypting Hcook: $Hcook");
		return (0,0,1,0,0);
	}
	my @splin=split(/:/,$Hcook_des);
    my $serv=pop @splin;
 	my $loc=pop @splin;
    my $crc=pop @splin;
    my $valid_date=pop @splin;
    my $temporal=shift @splin;
    my $code=join(':',@splin);

	my $gpoa_vdate=$valid_date;
	if ($code eq ""){
		$self->warning("Code parameter of Hcook is empty");
		return (0,0,1,0,0);
	}
	if ($loc ne $self->{'Loc'}){
		$self->warning("Location of Hcook is not valid:\n Loc of Hcook:#$loc#, Loc of this PoA:#$self->{'Loc'}#");
		return (0,0,1,0,0);
	}
	if ($serv ne $self->{'Serv'}){
		$self->warning("Service_ID of Hcook is not valid:\n Serv of Hcook:#$serv#, Serv of this PoA:#$self->{'Serv'}#");
		return (0,0,1,0,0);
	}
	if ($valid_date < $now){
		$self->warning("Hcook expired");
		return (0,0,1,0,0);
	}
	my $key=$code;
	if (defined $self->{AddressInToken} && $self->{AddressInToken} eq "1") {
		my ($cip,$d2);
		($key,$cip,$d2)=split /%/,$code;
	}
	for my $rejre (@{$self->{cookieRejects}}) {
		if ($key =~ /$rejre/) {
			$self->warning("$key matches rejection cookie filter $rejre");
			return (0,-1,1,0,0);
		}
	}
	$self->debug("Calling is_valid=$key,$crc,$gen_crc") if ($self->{'Debug'});
	my ($new_crc,$valid)=$self->is_valid($key,$crc,$gen_crc);
	$self->debug("Returning is_valid=$new_crc,$valid") if ($self->{'Debug'});
	if ($valid == 2){
		my $other=$self->get_crc($key);
		$self->warning("**COLLISION**: There is another Hcook with different CRC:\n Key in Hcook_DB=$key, Current CRC=#$crc#, CRC in Hcook_DB=#$other#");
		return (0,-1,1,0,0);
	}
	if ($valid == 3){
		$self->warning("There is no Hcook in Hcook_DB with key=#$key:$loc:$serv#");
		return (0,-1,1,0,0);
	}
	if ($valid == 1){
		$self->warning("Into CRC_Timeout");
		return ($code,1,1,0,$gpoa_vdate);
	}                                 
	if ($valid == 4){
		$self->warning("Error setting cookies in the browser. Too many tries");
		return (0,-1,1,0,0);
	}
	
	return ($code,$valid_date,$temporal,$new_crc,$gpoa_vdate);
}
												
sub gen_Lcook
{
	my $self=shift;
	my $code=shift;
	my $now=time;
	my $iv = pack("LLLL",rand(0xffffffff),rand(0xffffffff),rand(0xffffffff),rand(0xffffffff));
	my $key_aes=$self->{'Lkey'};
	my $dec_obj = new Crypt::CBC({ key=>$key_aes,cipher=>'Rijndael',iv=>$iv,regenerate_key=>0,padding=>'null',prepend_iv=>0});
	my $Lcook = encode_base64($iv.$dec_obj->encrypt("abcd:$now:$self->{'Loc'}:$self->{'Serv'}:$code"));
	$Lcook=uri_escape($Lcook);

	return "$self->{Lcook_Name}=$Lcook;path=$self->{'Loc'};domain=$self->{'Domain'}";
}

sub test_Lcook
{
	my $self=shift;
	my $Lcook=shift;
	
	if ($Lcook ne ""){
		my $enc_packet=decode_base64($Lcook);
    	my $iv = substr($enc_packet,0,16);
    	$enc_packet= substr($enc_packet,16);
     	my $key_aes=$self->{'Lkey'};
    	my $dec_obj = new Crypt::CBC({ key=>$key_aes,
				   cipher=>'Rijndael',
				   iv=>$iv,
                    regenerate_key=>0,
				   padding=>'null',
				   prepend_iv=>0
				   });
		my $Lcook_des=$dec_obj->decrypt($enc_packet);
		if ($Lcook_des ne ""){
			my @splin=split(/:/,$Lcook_des);
			shift @splin;
			my $tst=shift @splin;
			my $loc=shift @splin;
			my $serv=shift @splin;
			my $code=join(':',@splin);
			if ($code eq ""){
				$self->warning("Code parameter of Lcook is empty");
				return (0,"0");
			}
			if ($loc ne $self->{'Loc'}){
				$self->warning("Location parameter of Lcook is not valid:\n Loc of Lcook=#$loc#, Loc of PoA=#$self->{'Loc'}#");
				return (0,"$code");
			}
			if ($serv ne $self->{'Serv'}){
				$self->warning("Service_ID parameter of Lcook is not valid:\n Serv of Lcook=#$serv#, Serv of PoA=#$self->{'Serv'}#");
				return (0,"$code");
			}
			my $now=time;
			if ($tst + $self->{'Lcook_Timeout'} < $now){
				$self->warning("Lcook expired for $code");
				return (0,"$code");
			}
			my $key=$code;
			if (defined $self->{AddressInToken} &&
			    $self->{AddressInToken} eq "1") {
				my ($cip,$d2);
				($key,$cip,$d2)=split /%/,$code;
			}
			for my $rejre (@{$self->{cookieRejects}}) {
				if ($key =~ /$rejre/) {
					$self->warning("$key matches rejection cookie filter $rejre");
					return (0,"0");
				}
			}
			return (1,"$code");
        	}
		$self->warning("After decrypting Lcook is empty");
		return (0,"0");
	}
	else{
		$self->warning("Lcook is empty");
		return (0,"0");
	}	
}
	 
sub gen_logout_cookies
{

	my $self=shift;
	my $code=shift;
	
	
	my $crc=$self->del_Hcook($code);
	
	my $max_age="Tue, 01-Jan-1980 00:00:00 GMT";
	my $Hcook="$self->{Hcook_Name}=D;expires=$max_age;path=$self->{'Loc'};domain=$self->{'Domain'}";
	my $Lcook="$self->{Lcook_Name}=D;expires=$max_age;path=$self->{'Loc'};domain=$self->{'Domain'}";

	return ($Hcook,$Lcook);
}


#### DATABASE FUNCTIONS #####################

sub reg_Hcook
{
	my $self=shift;
	my ($base, %db);

	if ($self->{'registerKey'} eq "") {
		$self->error("Value of registerKey is null");
		return 0;
	}
	
	if(!($base = tie(%db, 'DB_File', $self->{'Hcook_DB'}, O_CREAT|O_RDWR, 0644, $DB_HASH))){
		$self->error("Opening Hcook DB: $!"); 
		return 0;
  }
  $db{$self->{'registerKey'}} = "$self->{registerVal}";
	untie %db;
	undef $base;
	$self->accesslog("Registering Hcook in Hcook_DB: ($self->{'registerKey'}),($self->{'registerVal'}:1)");
	return 1;
}

sub del_Hcook
{
	my $self=shift;
	
	my $code=$_[0];
	my $f_val=$_[1];

	my $crc=time;
	my $now=time;
	my($base, %db);
 
	if(!($base = tie(%db, 'DB_File', $self->{'Hcook_DB'}, O_CREAT|O_RDWR, 0644, $DB_HASH))){
		$self->error("Opening Hcook DB: $!");
		return 0;
	}
	delete $db{"$code:$self->{Loc}:$self->{Serv}"};
	untie %db;
	undef $base;
 
	$self->accesslog("Deleting Hcook from Hcook_DB: ($code:$self->{Loc}:$self->{Serv})");
	return 0;
}

sub is_valid
{
	my $self=shift;
	my $cod=shift;
	my $crc=shift;
	my $gen_crc=shift;
	my($base, %db, $now, $new_crc,$ret,$crc_last);

	my $locking = "write";
#	$locking = {
#		mode            => "read",
#		nonblocking     => 0,
#		lockfile_name   => "/path/to/shared.lock",
#		lockfile_mode   => 0600,
#	};

	my $now = time;
	srand;
	my $new_crc=int (10000 * rand ($now));

	my $key="$cod:$self->{Loc}:$self->{Serv}";
 
	if(!($base = tie(%db, 'DB_File::Lock', $self->{'Hcook_DB'}, O_CREAT|O_RDWR, 0644, $DB_HASH, $locking))){
		$self->error("Opening Hcook DB: $!");
		return 3;
	}
	my $res = undef;
	my $res = $db{"$key"};
  $self->debug("is_valid: DB Hcook=$res") if ($self->{'Debug'});

	if (defined $res) {
		my($crc_last,$crc_curr,$val_date,$created,$cont)=split(/:/,$res);
		if ($created + $self->{'CRC_Timeout'} > $now) {
      $self->debug("is_valid: into CRC_Tiemout= $created,$self->{'CRC_Timeout'},$now") if ($self->{'Debug'});
			$ret=1;
		}
		elsif ($crc_curr eq $crc){
			if ($gen_crc) {
      	$self->debug("is_valid: New CRC: $crc_curr:$new_crc:$val_date:$now:0") if ($self->{'Debug'});
      	$db{$key}="$crc_curr:$new_crc:$val_date:$now:0";
      }
			$ret=0;
		}
		elsif ($crc_last eq $crc) {
      $self->debug("is_valid: CRC=CRC_last") if ($self->{'Debug'});
			if ($gen_crc) {
				$cont++;
				$self->debug("is_valid: Increasing counter:$cont") if ($self->{'Debug'});
				if ($cont>$self->{'MxNonceErr'}){
					$ret=4;
				}else{
 					$self->debug("is_valid: New CRC: $crc_last:$new_crc:$val_date:$now:$cont") if ($self->{'Debug'});
					$db{$key}="$crc_last:$new_crc:$val_date:$now:$cont";
					$ret=0;
				}
			}else{
				$ret=0;
			}
		}else{
			$ret=2;
		}
	}else{
		$ret=3;
	}

	untie %db;
	undef $base;
	if ($gen_crc && ($ret==0)){
		return ($new_crc,$ret);
	}
	
	return (0,$ret);
}

sub get_crc {
	my $self=shift;
	my $cod=shift;
  my($base, %db);
 
	if(!($base = tie(%db, 'DB_File', $self->{'Hcook_DB'}, O_RDWR, 0644, $DB_HASH))){
		$self->error("Opening Hcook DB: $!");
		return "ERROR";
	}
	my $res = undef;
	my $locV=$db{"$cod:$self->{Loc}:$self->{Serv}"};
  if (defined $locV){
		$res = $db{"$cod:$self->{Loc}:$self->{Serv}"};
	}
	untie %db;
	undef $base;
 
	if (defined $res){
		my @campos=split(/:/,$res);
		return "$campos[0]:$campos[1]";
	}
	return "ERROR";
}
				     
sub read_key_file {
    my $file = shift;
    my $key_string;
    
    open(PRIV,$file) || die "$file: $!";
    read(PRIV,$key_string,-s PRIV); # Suck in the whole file
    close(PRIV);
    
    return $key_string;
}

sub rsaEncrypt {
	my $datarg = shift;
	my $keyFile = shift;
	
	# Get the byte size of data string 
	our $inputSize = 0;
	{
		use bytes;
		$inputSize = length($datarg);
	}

	# Generate a private key for encrypt and set the padding
	my $pkey = Crypt::OpenSSL::RSA->new_private_key(read_key_file($keyFile));
	$pkey->use_pkcs1_padding();

	# Get the output block maximun size
	### pkcs1_padding or sslv23_padding at most 11 bytes less than this size.
	# Get the size of block to encrypt and total number of blocks
	my $outputBlockSize = $pkey->size();
	my $inputBlockSize = $outputBlockSize - PADDINGSIZE;
	my $numBlocks = int($inputSize / $inputBlockSize);
	
	# Start to encrypt.
	my $blockCount = 0;
	my @cryptBuffer;

	while ($blockCount < $numBlocks){
		my $index = $blockCount * $inputBlockSize;
		my $block = substr($datarg, $index, $inputBlockSize);
		$cryptBuffer[$blockCount] = $pkey->private_encrypt($block);
		$blockCount++;
	}

	if ( ($inputSize % $inputBlockSize) != 0 ){
		$numBlocks++;
		my $index = $blockCount * $inputBlockSize;
		my $block = substr($datarg, $index, $inputBlockSize);
		$cryptBuffer[$blockCount] = $pkey->private_encrypt($block);
	}

	# Now joint the array with the blocks string encripted
	my $cryptData = join('',@cryptBuffer);
	my $base64CryptData = encode_base64($cryptData);
	
	# Return the encrypted, joined and base64 encode data string.
	return $base64CryptData;
}

#
###
#

sub rsaDecrypt{
	my $cryptData = shift;
	my $keyFile = shift;

	# Decode the base64 input string
	$cryptData = decode_base64($cryptData);

	# Get the byte size of data string 
	our $inputSize = 0;
	{
		use bytes;
		$inputSize = length($cryptData);
	}

	# Get the output block maximun size
  ## Non use the PADDINGSIZE to decrypt
	# Get the size of block to encrypt and total number of blocks
	my $pkey = Crypt::OpenSSL::RSA->new_public_key(read_key_file($keyFile));
	$pkey->use_pkcs1_padding();
	my $outputBlockSize = $pkey->size();
	my $inputBlockSize = $outputBlockSize;#- PADDINGSIZE;
	my $numBlocks = int($inputSize / $inputBlockSize);

	# Start to deencrypt.
	my $blockCount = 0;
	my @decryptBuffer = ();

	while ($blockCount < $numBlocks){
		my $index = $blockCount * $inputBlockSize;
		my $block = substr($cryptData, $index, $inputBlockSize);
		$decryptBuffer[$blockCount] = $pkey->public_decrypt($block);
		$blockCount++;
	}


	if ( ($inputSize % $inputBlockSize) != 0 ){
		$numBlocks++;
		my $index = $blockCount * $inputBlockSize;
		my $block = substr($cryptData, $index, $inputBlockSize);
		$decryptBuffer[$blockCount] = $pkey->public_decrypt($block);
	}
	# Now joint the array with the blocks string encripted

	my $decryptData = join('',@decryptBuffer);
	
	# Return the base64 dencode, decrypted and joined data string.
	return $decryptData;
}

1;

__END__
